@@ -7,9 +7,11 @@ import logging
 import os
 import pickletools
 from tarfile import TarError
+from tempfile import TemporaryDirectory
 from typing import IO, List, Optional, Set, Tuple
 import urllib.parse
 import zipfile
+from .relaxed_zipfile import RelaxedZipFile
 
 from .torch import (
     get_magic_number,
@@ -50,8 +52,9 @@ class ScanResult:
 
 
 class GenOpsError(Exception):
-    def __init__(self, msg: str):
+    def __init__(self, msg: str, globals: Optional[Set[Tuple[str, str]]]):
         self.msg = msg
+        self.globals = globals
         super().__init__()
 
     def __str__(self) -> str:
@@ -81,6 +84,16 @@ _safe_globals = {
         "IntStorage",
         "ByteStorage",
     },
+    "numpy": {
+        "dtype",
+        "ndarray",
+    },
+    "numpy._core.multiarray": {
+        "_reconstruct",
+    },
+    "numpy.core.multiarray": {
+        "_reconstruct",
+    },
     "torch._utils": {"_rebuild_tensor_v2"},
 }
 
@@ -103,25 +116,39 @@ _unsafe_globals = {
         "open",
         "breakpoint",
     },  # Pickle versions 3, 4 have those function under 'builtins'
-    "webbrowser": "*",  # Includes webbrowser.open()
-    "httplib": "*",  # Includes http.client.HTTPSConnection()
-    "requests.api": "*",
     "aiohttp.client": "*",
-    "os": "*",
+    "asyncio": "*",
+    "bdb": "*",
+    "commands": "*",  # Python 2 precursor to subprocess
+    "functools": "partial",  # functools.partial(os.system, "echo pwned")
+    "httplib": "*",  # Includes http.client.HTTPSConnection()
+    "numpy.testing._private.utils": "*",  # runstring() in this module is a synonym for exec()
     "nt": "*",  # Alias for 'os' on Windows. Includes os.system()
     "posix": "*",  # Alias for 'os' on Linux. Includes os.system()
+    "operator": "attrgetter",  # Ex of code execution: operator.attrgetter("system")(__import__("os"))("echo pwned")
+    "os": "*",
+    "requests.api": "*",
+    "runpy": "*",  # Includes runpy._run_code
+    "shutil": "*",
     "socket": "*",
+    "ssl": "*",  # DNS exfiltration via ssl.get_server_certificate()
     "subprocess": "*",
     "sys": "*",
+    "pdb": "*",
+    "pickle": "*",
+    "_pickle": "*",
+    "pip": "*",
+    "pydoc": "pipepager",  # pydoc.pipepager('help','echo pwned')
+    "timeit": "*",
+    "torch._inductor.codecache": "compile_file",  # compile_file('', '', ['sh', '-c','$(echo pwned)'])
+    "torch.serialization": "load",  # pickle could be used to load a different file
+    "venv": "*",
+    "webbrowser": "*",  # Includes webbrowser.open()
 }
 
 #
 # TODO: handle methods loading other Pickle files (either mark as suspicious, or follow calls to scan other files [preventing infinite loops])
 #
-# pickle.loads()
-# https://docs.python.org/3/library/pickle.html#pickle.loads
-# pickle.load()
-# https://docs.python.org/3/library/pickle.html#pickle.load
 # numpy.load()
 # https://numpy.org/doc/stable/reference/generated/numpy.load.html#numpy.load
 # numpy.ctypeslib.load_library()
@@ -136,9 +163,37 @@ _unsafe_globals = {
 # https://www.tensorflow.org/api_docs/python/tf/keras/models/load_model
 #
 
+_numpy_file_extensions = {".npy"}  # Note: .npz is handled as zip files
 _pytorch_file_extensions = {".bin", ".pt", ".pth", ".ckpt"}
-_pickle_file_extensions = {".pkl", ".pickle", ".joblib", ".dat", ".data", ".npy"}
-_zip_file_extensions = {".zip", ".npz"}
+_pickle_file_extensions = {".pkl", ".pickle", ".joblib", ".dat", ".data"}
+_zip_file_extensions = {".zip", ".npz", ".7z"}
+# Pickle files do not actually have magic bytes, but v2+ files
+# start with a PROTO (\x80) opcode followed by a byte with the protocol version
+_pickle_magic_bytes = {
+    b"\x80\x00",
+    b"\x80\x01",
+    b"\x80\x02",
+    b"\x80\x03",
+    b"\x80\x04",
+    b"\x80\x05",
+}
+_numpy_magic_bytes = b"\x93NUMPY"
+
+
+def _is_7z_file(f: IO[bytes]) -> bool:
+    read_bytes = []
+    start = f.tell()
+
+    byte = f.read(1)
+    while byte != b"":
+        read_bytes.append(byte)
+        if len(read_bytes) == 6:
+            break
+        byte = f.read(1)
+    f.seek(start)
+
+    local_header_magic_number = [b"7", b"z", b"\xbc", b"\xaf", b"\x27", b"\x1c"]
+    return read_bytes == local_header_magic_number
 
 
 def _http_get(url) -> bytes:
@@ -166,18 +221,20 @@ def _http_get(url) -> bytes:
 
 
 def _list_globals(data: IO[bytes], multiple_pickles=True) -> Set[Tuple[str, str]]:
-
     globals = set()
 
     memo = {}
     # Scan the data for pickle buffers, stopping when parsing fails or stops making progress
     last_byte = b"dummy"
+    parsing_pkl_error = None
     while last_byte != b"":
         # List opcodes
+        ops = []
         try:
-            ops = list(pickletools.genops(data))
+            for op in pickletools.genops(data):
+                ops.append(op)
         except Exception as e:
-            raise GenOpsError(str(e))
+            parsing_pkl_error = str(e)
         last_byte = data.read(1)
         data.seek(-1, 1)
 
@@ -189,13 +246,19 @@ def _list_globals(data: IO[bytes], multiple_pickles=True) -> Set[Tuple[str, str]
 
             if op_name == "MEMOIZE" and n > 0:
                 memo[len(memo)] = ops[n - 1][1]
-
-            if op_name == "GLOBAL":
+            elif op_name in ["PUT", "BINPUT", "LONG_BINPUT"] and n > 0:
+                memo[op_value] = ops[n - 1][1]
+            elif op_name in ("GLOBAL", "INST"):
                 globals.add(tuple(op_value.split(" ", 1)))
             elif op_name == "STACK_GLOBAL":
                 values = []
-                for offset in range(1, n):
-                    if ops[n - offset][0].name == "MEMOIZE":
+                for offset in range(1, n + 1):
+                    if ops[n - offset][0].name in [
+                        "MEMOIZE",
+                        "PUT",
+                        "BINPUT",
+                        "LONG_BINPUT",
+                    ]:
                         continue
                     if ops[n - offset][0].name in ["GET", "BINGET", "LONG_BINGET"]:
                         values.append(memo[int(ops[n - offset][1])])
@@ -204,6 +267,9 @@ def _list_globals(data: IO[bytes], multiple_pickles=True) -> Set[Tuple[str, str]
                         "UNICODE",
                         "BINUNICODE",
                         "BINUNICODE8",
+                        "STRING",
+                        "BINSTRING",
+                        "SHORT_BINSTRING",
                     ]:
                         _log.debug(
                             "Presence of non-string opcode, categorizing as an unknown dangerous import"
@@ -218,24 +284,27 @@ def _list_globals(data: IO[bytes], multiple_pickles=True) -> Set[Tuple[str, str]
                         f"Found {len(values)} values for STACK_GLOBAL at position {n} instead of 2."
                     )
                 globals.add((values[1], values[0]))
+
         if not multiple_pickles:
             break
 
+        if parsing_pkl_error is not None:
+            # XXX: given we can have multiple pickles in a file, we may have already successfully extracted globals from a valid pickle.
+            # Thus return the already found globals in the error & let the caller decide what to do.
+            # Additionally, we return the error at the end of the loop to scan imports in partially broken files,
+            # which can unpickle and be dangerous regardless of being valid pickle.
+            globals_opt = globals if len(globals) > 0 else None
+            raise GenOpsError(parsing_pkl_error, globals_opt)
+
     return globals
 
 
-def scan_pickle_bytes(data: IO[bytes], file_id, multiple_pickles=True) -> ScanResult:
-    """Disassemble a Pickle stream and report issues"""
-
+def _build_scan_result_from_raw_globals(
+    raw_globals: Set[Tuple[str, str]],
+    file_id,
+    scan_err=False,
+) -> ScanResult:
     globals = []
-    try:
-        raw_globals = _list_globals(data, multiple_pickles)
-    except GenOpsError as e:
-        _log.error(f"ERROR: parsing pickle in {file_id}: {e}")
-        return ScanResult(globals, scan_err=True)
-
-    _log.debug("Global imports in %s: %s", file_id, raw_globals)
-
     issues_count = 0
     for rg in raw_globals:
         g = Global(rg[0], rg[1], SafetyLevel.Dangerous)
@@ -261,28 +330,123 @@ def scan_pickle_bytes(data: IO[bytes], file_id, multiple_pickles=True) -> ScanRe
             g.safety = SafetyLevel.Suspicious
         globals.append(g)
 
-    return ScanResult(globals, 1, issues_count, 1 if issues_count > 0 else 0, False)
+    return ScanResult(globals, 1, issues_count, 1 if issues_count > 0 else 0, scan_err)
+
+
+def scan_pickle_bytes(data: IO[bytes], file_id, multiple_pickles=True) -> ScanResult:
+    """Disassemble a Pickle stream and report issues"""
+
+    try:
+        raw_globals = _list_globals(data, multiple_pickles)
+    except GenOpsError as e:
+        _log.error(f"ERROR: parsing pickle in {file_id}: {e}")
+        if e.globals is not None:
+            return _build_scan_result_from_raw_globals(
+                e.globals, file_id, scan_err=True
+            )
+        else:
+            return ScanResult([], scan_err=True)
+
+    _log.debug("Global imports in %s: %s", file_id, raw_globals)
+
+    return _build_scan_result_from_raw_globals(raw_globals, file_id)
+
+
+# XXX: it appears there is not way to get the byte stream for a given file within the 7z archive and thus forcing us to unzip to disk before scanning
+def scan_7z_bytes(data: IO[bytes], file_id) -> ScanResult:
+    try:
+        import py7zr
+    except ImportError:
+        raise Exception(
+            "py7zr is required to scan 7z archives, install picklescan using: 'pip install picklescan[7z]'"
+        )
+    result = ScanResult([])
+
+    with py7zr.SevenZipFile(data, mode="r") as archive:
+        file_names = archive.getnames()
+        targets = [f for f in file_names if f.endswith(tuple(_pickle_file_extensions))]
+        _log.debug("Files in 7z archive %s: %s", file_id, targets)
+        with TemporaryDirectory() as tmpdir:
+            archive.extract(path=tmpdir, targets=targets)
+            for file_name in targets:
+                file_path = os.path.join(tmpdir, file_name)
+                _log.debug("Scanning file %s in 7z archive %s", file_name, file_id)
+                if os.path.isfile(file_path):
+                    result.merge(scan_file_path(file_path))
+
+            return result
 
 
 def scan_zip_bytes(data: IO[bytes], file_id) -> ScanResult:
     result = ScanResult([])
 
-    with zipfile.ZipFile(data, "r") as zip:
+    with RelaxedZipFile(data, "r") as zip:
         file_names = zip.namelist()
-        _log.debug("Files in archive %s: %s", file_id, file_names)
+        _log.debug("Files in zip archive %s: %s", file_id, file_names)
         for file_name in file_names:
-            if os.path.splitext(file_name)[1] in _pickle_file_extensions:
-                _log.debug("Scanning file %s in zip archive %s", file_name, file_id)
+            try:
                 with zip.open(file_name, "r") as file:
-                    result.merge(scan_pickle_bytes(file, f"{file_id}:{file_name}"))
+                    magic_bytes = file.read(8)
+                file_ext = os.path.splitext(file_name)[1]
+
+                if file_ext in _pickle_file_extensions or any(
+                    magic_bytes.startswith(mn) for mn in _pickle_magic_bytes
+                ):
+                    _log.debug("Scanning file %s in zip archive %s", file_name, file_id)
+                    with zip.open(file_name, "r") as file:
+                        result.merge(scan_pickle_bytes(file, f"{file_id}:{file_name}"))
+
+                elif file_ext in _numpy_file_extensions or magic_bytes.startswith(
+                    _numpy_magic_bytes
+                ):
+                    _log.debug("Scanning file %s in zip archive %s", file_name, file_id)
+                    with zip.open(file_name, "r") as file:
+                        result.merge(scan_numpy(file, f"{file_id}:{file_name}"))
+            except (zipfile.BadZipFile, RuntimeError) as e:
+                # Log decompression issues (password protected, corrupted, etc.)
+                _log.warning(
+                    "Invalid file %s in zip archive %s: %s", file_name, file_id, str(e)
+                )
 
     return result
 
 
+def scan_numpy(data: IO[bytes], file_id) -> ScanResult:
+    # Delay import to avoid dependency on NumPy
+    import numpy as np
+
+    # Code to distinguish from NumPy binary files and pickles.
+    _ZIP_PREFIX = b"PK\x03\x04"
+    _ZIP_SUFFIX = b"PK\x05\x06"  # empty zip files start with this
+    N = len(np.lib.format.MAGIC_PREFIX)
+    magic = data.read(N)
+    # If the file size is less than N, we need to make sure not
+    # to seek past the beginning of the file
+    data.seek(-min(N, len(magic)), 1)  # back-up
+    if magic.startswith(_ZIP_PREFIX) or magic.startswith(_ZIP_SUFFIX):
+        # .npz file
+        raise ValueError(f".npz file not handled as zip file: {file_id}")
+    elif magic == np.lib.format.MAGIC_PREFIX:
+        # .npy file
+
+        version = np.lib.format.read_magic(data)
+        np.lib.format._check_version(version)
+        _, _, dtype = np.lib.format._read_array_header(data, version)
+
+        if dtype.hasobject:
+            return scan_pickle_bytes(data, file_id)
+        else:
+            return ScanResult([], 1)
+    else:
+        return scan_pickle_bytes(data, file_id)
+
+
 def scan_pytorch(data: IO[bytes], file_id) -> ScanResult:
     # new pytorch format
     if _is_zipfile(data):
         return scan_zip_bytes(data, file_id)
+    elif _is_7z_file(data):
+        return scan_7z_bytes(data, file_id)
     # old pytorch format
     else:
         scan_result = ScanResult([])
@@ -312,14 +476,17 @@ def scan_bytes(data: IO[bytes], file_id, file_ext: Optional[str] = None) -> Scan
         except InvalidMagicError as e:
             _log.error(f"ERROR: Invalid magic number for file {e}")
             return ScanResult([], scan_err=True)
+    elif file_ext is not None and file_ext in _numpy_file_extensions:
+        return scan_numpy(data, file_id)
     else:
         is_zip = zipfile.is_zipfile(data)
         data.seek(0)
-        return (
-            scan_zip_bytes(data, file_id)
-            if is_zip
-            else scan_pickle_bytes(data, file_id)
-        )
+        if is_zip:
+            return scan_zip_bytes(data, file_id)
+        elif _is_7z_file(data):
+            return scan_7z_bytes(data, file_id)
+        else:
+            return scan_pickle_bytes(data, file_id)
 
 
 def scan_huggingface_model(repo_id):
